/**
* IAX nagios plugin
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <arpa/inet.h>

#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>

/* IAX packet structure */
struct ast_iax2_full_hdr {
	unsigned short scallno;	/*!< Source call number -- high bit must be 1 */
	unsigned short dcallno;	/*!< Destination call number -- high bit is 1 if retransmission */
	unsigned int ts;		/*!< 32-bit timestamp in milliseconds (from 1st transmission) */
	unsigned char oseqno;	/*!< Packet number (outgoing) */
	unsigned char iseqno;	/*!< Packet number (next incoming expected) */
	unsigned char type;		/*!< Frame type */
	unsigned char csub;		/*!< Compressed subclass */
	unsigned char iedata[0];
} __attribute__ ((__packed__));

int crit_time=0,warn_time=0;
int rcvtimeo = 5000;

int iaxping(char * target_host)
{
	struct hostent  *target_hostent;
	unsigned long  target_hostaddr;
	struct sockaddr_in source_addr, target_addr;
	struct timeval tv;
	
	struct sockaddr from_address;
	socklen_t addr_length = sizeof(from_address);

	memset(&from_address, 0, sizeof(from_address));

	ssize_t retval;
	
	struct timeval tempo1, tempo2;
	long elapsed_mtime;    /* elapsed time in milliseconds */
	long elapsed_seconds;  /* diff between seconds counter */
	long elapsed_useconds; /* diff between microseconds counter */

	char receiveBuffer[1024];
	struct ast_iax2_full_hdr *pong,ping;
	
	memset(&ping, 0, sizeof(ping));
	ping.scallno=0x80; /*  IAX2 FULL */
	ping.type=0x06; /*  IAX type */
	ping.csub=0x1e; /*  POKE message */
	int iax_socket = socket(AF_INET, SOCK_DGRAM, 17);
	source_addr.sin_family = AF_INET;
	source_addr.sin_addr.s_addr=INADDR_ANY;
	
	tv.tv_sec  =  rcvtimeo / 1000;
        tv.tv_usec = (rcvtimeo % 1000) * 1000;
	
	/*  setting timeout */
	setsockopt(iax_socket, SOL_SOCKET, SO_RCVTIMEO,(struct timeval *)&tv,sizeof(struct timeval));
	
        if((target_hostent = gethostbyname(target_host)) == NULL)
        {
		printf( "SERVICE STATUS: ERROR Can't resolve address for %s\n", target_host);
                return 2;
        }

	target_hostaddr = *((unsigned long *)target_hostent->h_addr);
	target_addr.sin_family = AF_INET;
	target_addr.sin_addr.s_addr = target_hostaddr;
	target_addr.sin_port = htons(4569);
	retval = sendto(iax_socket, &ping, sizeof(ping), 0, (struct sockaddr *)&target_addr, sizeof(target_addr));
	gettimeofday(&tempo1, NULL);
	if(retval != sizeof(ping))
	{
		printf( "SERVICE STATUS: ERROR, Did not manage to send %ld bytes of data.\n",(long int)sizeof(ping));
		return 2;
	}

		retval = recvfrom(iax_socket, &receiveBuffer, 1024, 0, &from_address, &addr_length);
		if(retval != -1)
		{
			gettimeofday(&tempo2, NULL);
			elapsed_seconds  = tempo2.tv_sec  - tempo1.tv_sec;
			elapsed_useconds = tempo2.tv_usec - tempo1.tv_usec;
			elapsed_mtime = ((elapsed_seconds) * 1000 + elapsed_useconds/1000.0) + 0.5;
			
			pong = (struct ast_iax2_full_hdr *) receiveBuffer;
			
			if(pong->type==0x06 && pong->csub==0x03){ /* type=IAX, subclass=PONG */
				if(crit_time && elapsed_mtime>crit_time){
					fprintf(stdout, "SERVICE STATUS: CRITICAL, IAX PONG reply from: %s, in %ld ms\n", 
						inet_ntoa(((struct in_addr)((struct sockaddr_in *)&from_address)->sin_addr)),
						elapsed_mtime);
					return 1;
				}
				if(warn_time && elapsed_mtime>warn_time){
					fprintf(stdout, "SERVICE STATUS: WARNING, IAX PONG reply from: %s, in %ld ms\n", 
						inet_ntoa(((struct in_addr)((struct sockaddr_in *)&from_address)->sin_addr)),
						elapsed_mtime);
					return 1;
				}
				fprintf(stdout, "SERVICE STATUS: OK, IAX PONG reply from: %s, in %ld ms\n", 
					inet_ntoa(((struct in_addr)((struct sockaddr_in *)&from_address)->sin_addr)),
					elapsed_mtime);
				/*
				sending ACK to close connection 
				new version of Asterisk do no require this, but old 
				will resend PONG packets until timeout
				*/
				ping.csub=0x04; /* ACK */
				ping.iseqno=pong->iseqno;
				ping.dcallno=pong->scallno;
				sendto(iax_socket, &ping, sizeof(ping), 0, (struct sockaddr *)&target_addr, sizeof(target_addr));

				return 0;
			}
			else {
				fprintf(stdout, "SERVICE STATUS: ERROR, Wrong reply from: %s, in %ld ms\n", 
					inet_ntoa(((struct in_addr)((struct sockaddr_in *)&from_address)->sin_addr)),
					elapsed_mtime);
					return 2;
			}

		}
		else
		{	
			printf( "SERVICE STATUS: ERROR on recv(), errno %d\n",errno);
			return 2;
		}
		return 2; /* why we are there?? */
}

void printhelp(void){
	printf("IAX PING/PONG plugin for nagios, (c) Alex Samorukov \n"
		"Usage: check_iax [OPTION]\n"
		"\t-H <hostname> \t hostname or ip address of the IAX server\n"
		"\t-w <time> \t warning threshold, msec\n"
		"\t-c <time>  \t critical threshold, msec\n"
		"\t-t <timeout> \t recv() timeout, msec, default - %dms \n"
		"\t-h \t\t print this help\n"
		"Example: check_iax -H example.com\n",
		rcvtimeo);

}
int main(int argc, char * argv[])
{
	int c;
	char * hostaddr=NULL;
	
	while((c = getopt(argc, argv, "H:w:c:ht:")) != -1)
		switch (c)
        {
        case 'H':
		hostaddr=optarg;
		break;
        case 'h':
		printhelp();
		return 0;
		break;
        case 'w':
		warn_time=atoi(optarg);
		break;
	case 'c':
		crit_time=atoi(optarg);
		break;
	case 't':
		rcvtimeo=atoi(optarg);
		break;
	default:
		printhelp();
		return 1;		
        }
	
	if(argc < 2 || hostaddr==NULL)
	{
		printhelp();
		return 1;
	}
	return iaxping(hostaddr);
}
